/*
 * cmdufrag.c - IEEE1905 CMDU fragmentation and defragmentation handling
 *
 * Copyright (C) 2021 IOPSYS Software Solutions AB. All rights reserved.
 *
 * Author: anjan.chanda@iopsys.eu
 *
 * See LICENSE file for license related information.
 *
 */


#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>

#include "debug.h"
#include "timer.h"
#include "util.h"
#include "bufutil.h"
#include "1905_tlvs.h"
#include "cmdu.h"
#include "cmdufrag.h"



struct cmdu_buff *cmdu_fragment(uint8_t *data, int datalen)
{
	uint8_t *ptr = data, *pptr = NULL;
	struct cmdu_frag *frag = NULL;
	struct cmdu_buff *cmdu = NULL;
	int remlen = datalen;
	struct tlv *p = NULL;
	int len = 0;


	fprintf(stderr, "%s: >>>>>>>>\n", __func__);
	fprintf(stderr, "datalen = %d\n", datalen);
	/* bufprintf(data, datalen, "Data for fragment"); */

	if (datalen < FRAG_DATA_SIZE_TLV)
		return NULL;

	cmdu_for_each_tlv(p, data, remlen) {
		ptr = (uint8_t *)p;

		if (len + tlv_total_length(p) <= FRAG_DATA_SIZE_TLV) {
			len += tlv_total_length(p);
			dbg("%s: tlv-len = %d  seen-len = %d  remlen = %d\n",
			    __func__, tlv_total_length(p), len, remlen);
		} else {
			if (len == 0) {
				int idatalen = datalen;

				len = tlv_length(p);
				fprintf(stderr, "%s: tlv-len %d > %d\n", __func__, len, FRAG_DATA_SIZE);

				if (!cmdu) {
					cmdu = cmdu_alloc_frame(FRAG_DATA_SIZE_TLV);
					if (!cmdu) {
						fprintf(stderr, "%s: -ENOMEM\n", __func__);
						return NULL;
					}

					buf_put_be16(&data[1], FRAG_DATA_SIZE);	//update tlv length

					cmdu_put(cmdu, data, FRAG_DATA_SIZE_TLV);
					cmdu->num_frags = 0;
					fprintf(stderr, "%s: frag[0] len = %d\n", __func__, cmdu->datalen);
					//bufprintf(cmdu->data, cmdu->datalen, "Frag-0");
				}

				len -= FRAG_DATA_SIZE;
				ptr = data + FRAG_DATA_SIZE_TLV;
				idatalen -= FRAG_DATA_SIZE_TLV;
				fprintf(stderr, "%s: remain len = %d\n", __func__, len);

				while (len >= FRAG_DATA_SIZE) {
					frag = calloc(1, FRAG_DATA_SIZE_TLV + sizeof(*frag));
					if (!frag) {
						fprintf(stderr, "%s: -ENOMEM\n", __func__);
						return NULL;
					}

					frag->data = (uint8_t *)(frag + 1);
					frag->data[0] = p->type;
					buf_put_be16(&frag->data[1], FRAG_DATA_SIZE);

					memcpy(&frag->data[3], ptr, FRAG_DATA_SIZE);
					frag->len = FRAG_DATA_SIZE;
					list_add_tail(&frag->list, &cmdu->fraglist);
					cmdu->num_frags++;
					fprintf(stderr, "%s: frag[%d] len = %d\n", __func__,
						cmdu->num_frags, frag->len);
					//bufprintf(frag->data, frag->len, "Frag-x");

					len -= FRAG_DATA_SIZE;
					ptr += FRAG_DATA_SIZE;
					idatalen -= FRAG_DATA_SIZE;
				}

				/* handle residue */
				frag = calloc(1, FRAG_DATA_SIZE_TLV + sizeof(*frag));
				if (!frag) {
					fprintf(stderr, "%s: -ENOMEM\n", __func__);
					return NULL;
				}

				frag->data = (uint8_t *)(frag + 1);
				frag->data[0] = p->type;
				buf_put_be16(&frag->data[1], len);

				memcpy(&frag->data[3], ptr, len);
				frag->len = len + TLV_HLEN;

				idatalen -= len;
				fprintf(stderr, "Still left = %d\n", idatalen);
				memcpy(frag->data + frag->len, ptr + len, idatalen);
				frag->len += idatalen;

				list_add_tail(&frag->list, &cmdu->fraglist);
				cmdu->num_frags++;
				fprintf(stderr, "%s: frag[%d] len = %d\n", __func__,
					cmdu->num_frags, frag->len);
				//bufprintf(frag->data, frag->len, "Frag-x");

				goto print_frags;
			}

			if (!cmdu) {
				cmdu = cmdu_alloc_frame(FRAG_DATA_SIZE_TLV);
				if (!cmdu) {
					warn("%s: -ENOMEM\n", __func__);
					return NULL;
				}

				cmdu_put(cmdu, data, len);
				cmdu->num_frags = 0;
				dbg("%s: frag[0] len = %d\n", __func__, cmdu->datalen);
				/* bufprintf(cmdu->data, cmdu->datalen, "Frag-0"); */
			} else {
				frag = calloc(1, FRAG_DATA_SIZE_TLV + sizeof(*frag));
				if (!frag) {
					warn("%s: -ENOMEM\n", __func__);
					return NULL;
				}

				frag->data = (uint8_t *)(frag + 1);
				memcpy(frag->data, pptr, len);
				frag->len = len + TLV_HLEN;
				list_add_tail(&frag->list, &cmdu->fraglist);
				cmdu->num_frags++;
				dbg("%s: frag[%d] len = %d\n", __func__,
				    cmdu->num_frags, frag->len);
				/* bufprintf(frag->data, frag->len, "Frag-x"); */
			}

			len = tlv_total_length(p);
			pptr = ptr;
		}
	}

	/* residue */
	if (len > 0) {
		if (!cmdu) {
			cmdu = cmdu_alloc_frame(FRAG_DATA_SIZE_TLV);
			if (!cmdu) {
				warn("%s: -ENOMEM\n", __func__);
				return NULL;
			}

			cmdu_put(cmdu, data, len);
			cmdu->num_frags = 0;
			dbg("%s: frag[0] len = %d\n", __func__, cmdu->datalen);
			/* bufprintf(cmdu->data, cmdu->datalen, "Frag-0"); */
		} else {
			frag = calloc(1, FRAG_DATA_SIZE_TLV + sizeof(*frag));
			if (!frag) {
				warn("%s: -ENOMEM\n", __func__);
				return NULL;
			}

			frag->data = (uint8_t *)(frag + 1);
			memcpy(frag->data, pptr, len);
			frag->len = len;
			list_add_tail(&frag->list, &cmdu->fraglist);
			cmdu->num_frags++;
			dbg("%s: frag[%d] len = %d\n", __func__, cmdu->num_frags,
			    frag->len);
			/* bufprintf(frag->data, frag->len, "Frag-x"); */
		}
	}

print_frags:
#ifdef I1905_DEBUG
{
	int i;

	fprintf(stderr, "\ncmdu fragments\n");
	fprintf(stderr, "CMDU: len = %d\n", cmdu->datalen);
	bufprintf(cmdu->data, cmdu->datalen, "Fragment-0");

	list_for_each_entry(frag, &cmdu->fraglist, list) {
		fprintf(stderr, "FRAGMENT: fraglen = %d\n", frag->len);
		bufprintf(frag->data, frag->len, "Fragment-i");
	}
}
#endif

	return cmdu;
}

static struct cmdu_frag_rx *cmdufrag_lookup(void *rxfq, uint16_t type,
					    uint16_t mid, uint8_t fid,
					    uint8_t *origin)
{
	struct cmdufrag_queue *q = (struct cmdufrag_queue *)rxfq;
	int idx = cmdu_frag_hash(type, mid, origin);
	struct cmdu_frag_rx *frag = NULL;


	pthread_mutex_lock(&q->qlock);
	hlist_for_each_entry(frag, &q->table[idx], hlist) {
		if (frag->type == type && frag->mid == mid &&
		    !memcmp(frag->origin, origin, 6)) {
			if (frag->fid == fid) {
				pthread_mutex_unlock(&q->qlock);
				return frag;
			}
		}
	}
	pthread_mutex_unlock(&q->qlock);

	return NULL;
}

static void cmdu_fragqueue_delete_chain(struct cmdu_frag_rx *frag)
{
	struct cmdu_frag_rx *ee = NULL, *e;

	if (!frag)
		return;

	for (e = frag; e; e = e->next) {
		if (ee) {
			printf("freeing ee ..fid = %d ....\n", ee->fid);
			free(ee);
		}
		ee = e;
	}

	if (ee) {
		printf("freeing ee ..fid = %d ...\n", ee->fid);
		free(ee);
	}
}

static void cmdu_fragqueue_free_entry(struct cmdu_frag_rx *frag)
{
	if (frag)
		free(frag);
}

static struct cmdu_frag_rx *cmdu_create_rxfrag(struct cmdu_buff *cmdu,
					       uint32_t timeout)
{
	struct cmdu_frag_rx *frag;
	struct timeval tsp = { 0 };

	frag = calloc(1, sizeof(*frag));
	if (!frag) {
		fprintf(stderr, "calloc failed. err = NOMEM\n");
		return NULL;
	}

	frag->cmdu = cmdu;
	frag->type = cmdu_get_type(cmdu);
	frag->mid = cmdu_get_mid(cmdu);
	frag->fid = cmdu_get_fid(cmdu);
	frag->last_frag = IS_CMDU_LAST_FRAGMENT(cmdu->cdata) ? true : false;
	memcpy(frag->origin, cmdu_get_origin(cmdu), 6);
	frag->next = NULL;
	frag->last = frag;
	frag->tlen = cmdu->datalen;	//0;
	frag->numfrags = 1;
	gettimeofday(&tsp, NULL);
	frag->ageing_time = timeout;
	timeradd_msecs(&tsp, frag->ageing_time, &frag->ageing_tmo);
	frag->ageing_tmo.tv_usec = (frag->ageing_tmo.tv_usec / 1000) * 1000;
	fprintf(stderr,
		"CREATE frag: type = 0x%04x  mid = %hu (%d) datalen = %d  origin = " MACFMT " timeout = { %u (%jd:%jd) }\n",
		cmdu_get_type(cmdu),
		cmdu_get_mid(cmdu),
		cmdu_get_fid(cmdu),
		frag->tlen,
		MAC2STR(cmdu_get_origin(cmdu)),
		frag->ageing_time,
		(intmax_t) frag->ageing_tmo.tv_sec,
		(intmax_t) frag->ageing_tmo.tv_usec / 1000);

	return frag;
}

int cmdufrag_queue_enqueue(void *rxfq, struct cmdu_buff *cmdu, uint32_t timeout)
{
	struct cmdufrag_queue *q = (struct cmdufrag_queue *)rxfq;
	struct cmdu_frag_rx *frag = NULL;
	uint8_t *origin;
	uint16_t type;
	uint16_t mid;
	uint8_t fid;


	type = cmdu_get_type(cmdu);
	mid = cmdu_get_mid(cmdu);
	fid = cmdu_get_fid(cmdu);
	origin = cmdu_get_origin(cmdu);

	frag = cmdufrag_lookup(rxfq, type, mid, fid, origin);
	if (frag) {
		fprintf(stderr,
			"DROP duplicate: type = 0x%04x mid = %hu fid = %d origin = " MACFMT "\n",
			type, mid, fid, MAC2STR(origin));

		return -1;
	}

	frag = cmdu_create_rxfrag(cmdu, timeout);
	if (frag) {
		int idx = cmdu_frag_hash(type, mid, origin);

		pthread_mutex_lock(&q->qlock);
		q->pending_cnt++;
		pthread_mutex_unlock(&q->qlock);

		fprintf(stderr,
			"ENQ: type = 0x%04x  mid = %hu fid = %d  origin = " MACFMT "\n",
			type, mid, fid, MAC2STR(origin));

		if (fid > 0) {
			struct cmdu_frag_rx *firstfrag = NULL;

			firstfrag = cmdufrag_lookup(rxfq, type, mid, 0, origin);
			if (!firstfrag) {
				fprintf(stderr,
					"First fragment missing for mid = %hu\n", mid);
				cmdu_fragqueue_free_entry(frag);
				return -1;
			}

			pthread_mutex_lock(&q->qlock);
			firstfrag->last->next = frag;
			firstfrag->last = frag;
			firstfrag->tlen += frag->cmdu->datalen;
			firstfrag->numfrags++;
			fprintf(stderr, "%s: tlen = %d   numfrags = %d\n", __func__,
				firstfrag->tlen, firstfrag->numfrags);

			/* do not ageout fragments other than the first.
			 * If the first one ages-out, then all the related
			 * fragments will be cleaned up.
			 */
			pthread_mutex_unlock(&q->qlock);
			return 0;
		}

		pthread_mutex_lock(&q->qlock);
		hlist_add_head(&frag->hlist, &q->table[idx]);

		if (timer_pending(&q->ageing_timer)) {
			if (timercmp(&q->next_tmo, &frag->ageing_tmo, >)) {
				q->next_tmo.tv_sec = frag->ageing_tmo.tv_sec;
				q->next_tmo.tv_usec = frag->ageing_tmo.tv_usec;

				timer_set(&q->ageing_timer, frag->ageing_time);
			}
		} else {
			q->next_tmo.tv_sec = frag->ageing_tmo.tv_sec;
			q->next_tmo.tv_usec = frag->ageing_tmo.tv_usec;
			timer_set(&q->ageing_timer, frag->ageing_time);
		}

		pthread_mutex_unlock(&q->qlock);
		return 0;
	}

	return -1;
}

static void cmdu_frag_ageout(struct cmdufrag_queue *st, struct hlist_head *head,
			     struct timeval *min_next_tmo)
{
	struct cmdu_frag_rx *frag;
	struct hlist_node *tmp;
	struct timeval now = { 0 };


	gettimeofday(&now, NULL);
	now.tv_usec = (now.tv_usec / 1000) * 1000;

	hlist_for_each_entry_safe(frag, tmp, head, hlist) {
		if (timercmp(&frag->ageing_tmo, &now, <=)) {
			st->pending_cnt--;
			hlist_del(&frag->hlist, head);
			fprintf(stderr, "Fragments from " MACFMT " aged out.\n",
				MAC2STR(frag->origin));
			cmdu_fragqueue_delete_chain(frag);
		} else {
			struct timeval new_next_tmo = { 0 };

			timersub(&frag->ageing_tmo, &now, &new_next_tmo);
			if (!timercmp(min_next_tmo, &new_next_tmo, <)) {
				min_next_tmo->tv_sec = new_next_tmo.tv_sec;
				min_next_tmo->tv_usec = new_next_tmo.tv_usec;
			}
		}
	}
}

static void cmdu_fragqueue_ageing_timer_run(atimer_t *t)
{
	struct cmdufrag_queue *st = container_of(t, struct cmdufrag_queue, ageing_timer);
	struct timeval min_next_tmo = { .tv_sec = 999999 };
	int remain_cnt = 0;
	struct timeval nu;
	int i;

	gettimeofday(&nu, NULL);

	fprintf(stderr, "\n Timer now = %jd.%jd   cnt = %d\n",
		(intmax_t) nu.tv_sec, (intmax_t) nu.tv_usec, st->pending_cnt);

	for (i = 0; i < NUM_FRAGMENTS; i++) {
		if (hlist_empty(&st->table[i]))
			continue;

		cmdu_frag_ageout(st, &st->table[i], &min_next_tmo);
	}

	remain_cnt = st->pending_cnt;
	st->next_tmo.tv_sec = min_next_tmo.tv_sec;
	st->next_tmo.tv_usec = min_next_tmo.tv_usec;

	if (remain_cnt) {
		uint32_t tmo_msecs = st->next_tmo.tv_sec * 1000 +
					st->next_tmo.tv_usec / 1000;

		if (tmo_msecs > 0)
			timer_set(&st->ageing_timer, tmo_msecs);
	}
}

void free_rxfrag(struct cmdu_frag_rx *e) {
	if (e->next) {
		free_rxfrag(e->next);
	}


	fprintf(stderr, "freeing fid = %d\n", e->fid);
	cmdu_free(e->cmdu);
	free(e);
}

struct cmdu_buff *cmdu_defrag(void *rxfq, struct cmdu_buff *lastfrag)
{
	struct cmdufrag_queue *q = (struct cmdufrag_queue *)rxfq;
	struct cmdu_frag_rx *e;
	struct cmdu_frag_rx *frag = NULL;
	struct cmdu_buff *cmdu = NULL;
	struct hlist_node *tmp;
	uint32_t fidsum = 0;
	bool is_lastfrag;
	uint8_t *origin;
	uint16_t type;
	uint16_t mid;
	uint8_t fid;
	int idx;


	fprintf(stderr, "%s() >>>>>>>>\n", __func__);
	if (!lastfrag || !lastfrag->cdata)
		return NULL;

	type = cmdu_get_type(lastfrag);
	mid = cmdu_get_mid(lastfrag);
	fid = cmdu_get_fid(lastfrag);
	is_lastfrag = IS_CMDU_LAST_FRAGMENT(lastfrag->cdata);
	origin = cmdu_get_origin(lastfrag);

	if (!is_lastfrag)
		return NULL;

	idx = cmdu_frag_hash(type, mid, origin);

	pthread_mutex_lock(&q->qlock);
	hlist_for_each_entry_safe(frag, tmp, &q->table[idx], hlist) {
		if (frag->type == type && frag->mid == mid &&
			!memcmp(frag->origin, origin, 6)) {

			fprintf(stderr, "DEFRAG: type: 0x%04x mid = %hu  fid = %d " \
					"numfrags: %d\n", frag->type, frag->mid,
					frag->fid, frag->numfrags);

			hlist_del(&frag->hlist, &q->table[idx]);
			q->pending_cnt -= frag->numfrags;
			break;
		}
	}
	pthread_mutex_unlock(&q->qlock);

	if (!frag)
		return NULL;

	/* alloc unfragmented cmdu */
	cmdu = cmdu_alloc_frame(frag->tlen + 3);  /* including EOM if missing */
	if (!cmdu) {
		fprintf(stderr, "-ENOMEM\n");
		return NULL;
	}

	fprintf(stderr, "%s: reassembled CMDU datalen = %d\n", __func__, frag->tlen);
	cmdu_set_type(cmdu, type);
	cmdu_set_mid(cmdu, mid);
	memcpy(cmdu->origin, origin, 6);
	memcpy(cmdu->dev_macaddr, lastfrag->dev_macaddr, 6);
	strncpy(cmdu->dev_ifname, lastfrag->dev_ifname, 15);
	CMDU_SET_LAST_FRAGMENT(cmdu->cdata);

	for (e = frag; e; e = e->next) {
		fidsum += e->fid;
		memcpy(cmdu->tail, e->cmdu->data, e->cmdu->datalen);
		cmdu->datalen += e->cmdu->datalen;
		cmdu->tail += e->cmdu->datalen;
	}
	fprintf(stderr, "%s: reassembled CMDU datalen (%d), expected len (%d)\n", __func__, cmdu->datalen, frag->tlen);

	free_rxfrag(frag);

	fprintf(stderr, "frags sum = %d\n", fidsum);
	if (fid * (fid + 1) != 2 * fidsum) {
		fprintf(stderr, "Defrag Failure!\n");
		cmdu_free(cmdu);
		return NULL;
	}

	bufprintf(cmdu->data, cmdu->datalen, "=== DEFRAG CMDU ===");

	return cmdu;
}

int cmdufrag_queue_init(void *rxfq)
{
	struct cmdufrag_queue *q = (struct cmdufrag_queue *)rxfq;

	memset(q, 0, sizeof(*q));
	pthread_mutex_init(&q->qlock, NULL);
	timer_init(&q->ageing_timer, cmdu_fragqueue_ageing_timer_run);

	return 0;
}

void cmdufrag_queue_flush(void *rxfq)
{
	struct cmdufrag_queue *q = (struct cmdufrag_queue *)rxfq;
	struct cmdu_frag_rx *msg = NULL;
	int idx = 0;

	pthread_mutex_lock(&q->qlock);
	for (idx = 0; idx < NUM_FRAGMENTS; idx++) {
		hlist_for_each_entry(msg, &q->table[idx], hlist)
			cmdu_fragqueue_delete_chain(msg);

		q->table[idx].first = NULL;
	}

	q->pending_cnt = 0;
	pthread_mutex_unlock(&q->qlock);
}

void cmdufrag_queue_free(void *rxfq)
{
	struct cmdufrag_queue *q = (struct cmdufrag_queue *)rxfq;

	cmdufrag_queue_flush(q);
	timer_del(&q->ageing_timer);
	pthread_mutex_destroy(&q->qlock);
}
